﻿using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using static WSCF.Reflect_OBJ_INTERFACE;

namespace WSCF
{
    /// <summary>
    /// obj 服务代理;
    /// 1.如果代理的对象为 INotifyPropertyChanged, 对象内的 property 改变，能推送给客户端( [JsonIgnore] 除外);
    /// 2.代理的对象不为INotifyPropertyChanged, 但 对象内的 property 
    ///     a.是 [PropertyPush], 
    ///     b.且是 INotifyPropertyChanged, 
    ///     c.且 只能是get,不能有set,
    /// 那这个 property 内的 子property 改变，也会推送给 客户端 且 子property 是 [PropertyPush]，会继续向下全部注册 PropertyChanged,
    /// 也推送给 客户端;
    /// 3. 客户端 可以通过 Call(.....) 调用 服务 的function,  function返回的内容，需要通过[Call(...)] 定义; 且function 必须是
    /// void function(object param1, object param2,object paramN, AsyncCBHandler asyncDelegate, object asyncContext);
    /// 必须有参数 AsyncCBHandler asyncDelegate, object asyncContext， 名字不能改。 前面的 param1~paramN， 多少个都行
    /// 4. 服务中的event, 通过 [Push(...)] 可以推送给客户端， 但event 只能是 EventHandler
    /// [Push(typeof(BulkDBTempFrameChangedEventArgs))]
    ///  public event EventHandler TempFrameChanged;
    ///  
    /// </summary>
    public class WsServiceProxy : IDisposable
    {
        static NLog.Logger logger = NLog.LogManager.GetCurrentClassLogger();
        #region Core
        /// <summary>
        /// 对象的接口类型
        /// </summary>
        Type interfaceType;
        /// <summary>
        /// 代理对象
        /// </summary>
        object obj;
        /// <summary>
        /// 忽略设置
        /// </summary>
        public bool ignoreSet = false;
        /// <summary>
        /// 对象的全部event
        /// </summary>
        List<AnyEvent> anyEvents = new List<AnyEvent>();
        /// <summary>
        /// [PropertyPush] 标签树 
        /// </summary>
        SubPropertyNode rootNode;
        /// <summary>
        /// interfaceType 及其父类 的全部属性
        /// </summary>
        List<string> properties = new List<string>();
        #endregion
        public bool isAlive = false;

        //向客户端 回复数据
        public Action<string> SendEx;
        /// <summary>
        /// 服务在 服务器中
        /// </summary>
        public bool IsServiceInServer = true;
        /// <summary>
        /// 服务名称
        /// </summary>
        public string ServiceName;
        public WsServiceProxy()
        {

        }
        public void SetSendEx(Action<string> sendEx, string serviceName) 
        {
            this.ServiceName = serviceName;
            this.SendEx = sendEx;
        }
        #region Core
        public void Init(Type interfaceType, object obj)
        {
            this.interfaceType = interfaceType;
            this.obj = obj;

            rootNode = new SubPropertyNode
            {
                Obj = obj,
                InterfaceType = interfaceType,
                SubPropertyChanged = Sub_PropertyChanged
            };

            //注册 obj 的PropertyChanged 事件，获取 interfaceType 全部属性名称，包括它的父类的全部属性名称
            COMMON.InitPropertyChanged(interfaceType, obj, rootNode, properties);

            //处理[PropertyPush]
            COMMON.InitPropertyPush(rootNode);

            //处理[Push]
            InitEventPush();
        }

        void InitEventPush()
        {
            var interfaceTypes = new List<Type>();

            interfaceTypes.Add(this.interfaceType);
            interfaceTypes.AddRange(this.interfaceType.GetInterfaces());


            var eventInfos = new List<EventInfo>();
            foreach (var ifaceType in interfaceTypes)
            {
                eventInfos.AddRange(ifaceType.GetEvents());
            }

            foreach (var eventInfo in eventInfos)
            {
                var pushAttribute = eventInfo.GetCustomAttribute<PushAttribute>();
                if (pushAttribute == null)
                    continue;

                if (anyEvents.Any(ae => ae.eventName == eventInfo.Name))
                    continue;//已经添加了

                var anyEvent = new AnyEvent()
                {
                    eventName = eventInfo.Name,
                    PushObjInfoEx = (rd) => reponse_PushObjInfoEx(rd, true)
                };
                //这个事件需要推送
                eventInfo.AddEventHandler(obj, anyEvent.eventHandler);
                anyEvents.Add(anyEvent);

            }
        }
        void Sub_PropertyChanged(SubPropertyNode node, PropertyChangedEventArgs e)
        {
            var rData = COMMON.Sub_PropertyChanged(node, e, rootNode, properties);
            if (rData == null)
                return;

            reponse_PushObjInfoEx(rData, false);
        }

        ReflectData request_CALL_GetAllProperties()
        {

            var jObject = new JObject();

            var type = obj.GetType();

            foreach (var propertyName in properties)
            {
                var propertyInfo = type.GetProperty(propertyName);
                var v = propertyInfo.GetValue(obj);
                if (v == null)
                    jObject.Add(propertyInfo.Name, null);
                else
                    jObject.Add(propertyInfo.Name, JToken.FromObject(v));
            }

            var rData = new ReflectData
            {
                data = jObject
            };
            return rData;
        }

        void request_CALL_SetProperty(ReflectData rData)
        {
            //ignoreSet = true;

            //sub property
            var node = COMMON.FindNode(rootNode, rData.name);
            if (node == null)
            {
                //异常
                //客户端乱发过来
                return;
            }

            string json = rData.data.ToString();
            JsonConvert.PopulateObject(json, node.Obj);

            //ignoreSet = false;
        }

        void request_CALL_MethodInvoke(ReflectData rData, string guid, Reflect_Proxy proxy)
        {
            var type = obj.GetType();
            var paramNames_req = rData.data.Children().OfType<JProperty>().Select(p => p.Name);
            MethodInfo methodInfo = GetMethodInfo(type, rData.name, paramNames_req);
            if (methodInfo == null)
            {
                //不能找到，
                throw new Exception($"程序写错了, 不能找到 rData.name={rData.name} 或者 参数名称不对，没法完全匹配");
            }

            var parameterInfos = methodInfo.GetParameters();
            object[] parameters = new object[parameterInfos.Count()];
            for (int i = 0; i < parameters.Count(); i++)
            {
                var ptype = parameterInfos[i].ParameterType;
                var pname = parameterInfos[i].Name;
                if (string.Compare(pname, Reflect_OBJ_INTERFACE.paramName_asyncDelegate, true) == 0)
                {
                    parameters[i] = new AsyncCBHandler(asyncDelegate);
                }
                else if (string.Compare(pname, Reflect_OBJ_INTERFACE.paramName_asyncContext, true) == 0)
                {
                    parameters[i] = new CC() { proxy = proxy, guid = guid, methodName = rData.name };
                }
                else
                {
                    parameters[i] = rData.data[pname].ToObject(ptype);
                }
            }
            methodInfo.Invoke(obj, parameters);
        }

        void asyncDelegate(object asyncContext, object retData)
        {
            var cc = (CC)asyncContext;

            var rData = new ReflectData()
            {
                name = cc.methodName,
                data = retData == null ? null : JToken.FromObject(retData)
            };

            reponse_CALL_MethodInvoke(rData, cc.guid);
        }

        MethodInfo GetMethodInfo(Type type, string name, IEnumerable<string> parameterNames)
        {

            var methodInfos = from mi in type.GetMethods() where mi.Name == name select mi;
            if (methodInfos.Count() == 0)
                return null;
            if (methodInfos.Count() == 1)
                return methodInfos.First();

            //必须完全匹配
            foreach (var methodInfo in methodInfos)
            {
                var parameterInfos = methodInfo.GetParameters();
                //全部参数名称
                var names = parameterInfos.Select(pi => pi.Name).ToList();
                //删除掉 asyncDelegate，asyncContext
                names.Remove(Reflect_OBJ_INTERFACE.paramName_asyncDelegate);
                names.Remove(Reflect_OBJ_INTERFACE.paramName_asyncContext);

                var names_req = parameterNames;

                if (names.Count() != names_req.Count())
                    continue;//数量不一致，肯定不同

                var sames = names_req.Intersect(names);

                if (sames.Count() != names_req.Count())
                    continue;// names 与 names_req 的交集数量与names_req不一样，肯定不同

                //就是它
                return methodInfo;

            }
            return null;
        }

        public void Dispose()
        {
            if (typeof(INotifyPropertyChanged).IsAssignableFrom(interfaceType))
            {
                //继承了INotifyPropertyChanged
                ((INotifyPropertyChanged)(this.obj)).PropertyChanged -= rootNode.PropertyChanged;
            }

            //释放subProperties
            COMMON.NodeDispose(rootNode);
        }
        #endregion


        void reponse_PushObjInfoEx(ReflectData rData, bool isEvent)
        {
            //数据推送！！！
            //再嵌套
            var pkgName = isEvent ? PkgName.PUSH_Event : PkgName.PUSH_PropertyChanged;

            var pkgData = new PkgData(
                ServiceName,
                pkgName,
                Guid.NewGuid().ToString(),
                JToken.FromObject(rData));

            string json = JsonConvert.SerializeObject(pkgData);
            SendEx?.Invoke(json);
        }

        void reponse_CALL_MethodInvoke(ReflectData rData, string guid)
        {

            var pkgData = new PkgData(
                ServiceName,
                PkgName.CALL_MethodInvoke,
                guid,
                JToken.FromObject(rData));

            string json = JsonConvert.SerializeObject(pkgData);
            SendEx?.Invoke(json);
        }

        void reponse_CALL_GetAllProperties(ReflectData rData, string guid)
        {

            var pkgData = new PkgData(
                ServiceName,
                PkgName.CALL_GetAllProperties,
                guid,
                JToken.FromObject(rData));

            string json = JsonConvert.SerializeObject(pkgData);
            SendEx?.Invoke(json);
        }



        #region 被 WsService 调用
        public void OnMessage(Reflect_Proxy proxy, PkgData pkgData)
        {
            var pkgName = Enum.Parse(typeof(PkgName), pkgData.name);
            switch (pkgName)
            {
                case PkgName.CALL_GetAllProperties:
                    {
                        var rData = request_CALL_GetAllProperties();

                        reponse_CALL_GetAllProperties(rData, pkgData.guid);
                    }
                    break;
                case PkgName.CALL_SetProperty:
                    {
                        //ignoreSet = true;
                        var rData = pkgData.data.ToObject<ReflectData>();

                        request_CALL_SetProperty(rData);
                    }
                    break;
                case PkgName.CALL_MethodInvoke:
                    {
                        var rData = pkgData.data.ToObject<ReflectData>();

                        request_CALL_MethodInvoke(rData, pkgData.guid, proxy);

                    }
                    break;
            }
        }




        public void OnOpen()
        {
            isAlive = true;
            AfterOpen?.Invoke(this);
        }

        public Action<WsServiceProxy> AfterOpen;

        public void OnClose()
        {
            isAlive = false;
             Dispose();
        }

        #endregion
    }
    /// <summary>
    /// 事件推送包装
    /// </summary>
    class AnyEvent
    {
        /// <summary>
        /// 事件名称
        /// </summary>
        public string eventName;
        /// <summary>
        /// json包装后的ReflectData 推送
        /// </summary>
        public Action<ReflectData> PushObjInfoEx;
        /// <summary>
        /// 给 EventInfo 调用
        /// </summary>
        public EventHandler eventHandler;
        public AnyEvent()
        {
            eventHandler = new EventHandler(Obj_AnyEvent);
        }
        void Obj_AnyEvent(object sender, EventArgs e)
        {
            var rData = new ReflectData()
            {
                name = eventName,
                data = JObject.FromObject(e)
            };
            PushObjInfoEx?.Invoke(rData);
        }
    }

    public class CC
    {
        /// <summary>
        /// 连接
        /// </summary>
        public Reflect_Proxy proxy;


        public string guid;
        /// <summary>
        /// CALL_MethodInvoke 时使用
        /// </summary>
        public string methodName;

    }
}
